@@ -1,5 +1,5 @@
 /*  xcowsay.c -- Cute talking cow for GTK+.
- *  Copyright (C) 2008  Nick Gasson
+ *  Copyright (C) 2008-2010  Nick Gasson
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -15,6 +15,7 @@
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#define _GNU_SOURCE
 #include <stdio.h>
 #include <stdlib.h>
 #include <time.h>
@@ -22,6 +23,11 @@
 #include <assert.h>
 #include <string.h>
 
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <unistd.h>
+#include <errno.h>
+
 #include "display_cow.h"
 #include "settings.h"
 #include "xcowsayd.h"
@@ -43,6 +49,8 @@
 #define DEF_COW_SIZE      "med"
 #define DEF_IMAGE_BASE    "cow"
 #define DEF_DREAM_TIME    10000
+#define DEF_ALT_IMAGE     ""
+#define DEF_BUBBLE_X      5  // Distance from cow to bubble
 
 #define MAX_STDIN 4096   // Maximum chars to read from stdin
 
@@ -60,6 +68,11 @@ static struct option long_options[] = {
    {"cow-size", required_argument, 0, 'c'},
    {"reading-speed", required_argument, 0, 'r'},
    {"daemon", no_argument, &daemon_flag, 1},
+   {"image", required_argument, 0, 'i'},
+   {"monitor", required_argument, 0, 'm'},
+   {"bubble-at", required_argument, 0, 'b'},
+   {"at", required_argument, 0, 'a'},
+   {"config", required_argument, 0, 'o'},
    {"debug", no_argument, &debug, 1},
    {0, 0, 0, 0}
 };
@@ -80,49 +93,59 @@ static void read_from_stdin(cowmode_t mode)
 
 static void usage()
 {
-   printf
-      ("%s: xcowsay [OPTION]... [MESSAGE]...\n"
-       "%s\n\n"
-       "%s:\n"
-       " -h, --help\t\t%s\n"
-       " -v, --version\t\t%s\n"
-       " -t, --time=SECONDS\t%s\n"
-       " -r, --reading-speed=N\t%s\n"
-       " -f, --font=FONT\t%s\n"
-       " -d, --dream=FILE\t%s\n"
-       "     --think\t\t%s\n"
-       "     --daemon\t\t%s\n"
-       "     --cow-size=SIZE\t%s\n"
-       "     --debug\t\t%s\n\n"
-       "%s\n\n"
-       "%s\n\n"
-       "%s\n",
-       i18n("Usage"),
-       i18n("Display a cow on your desktop with MESSAGE or standard input."),
-       i18n("Options"),
-       i18n("Display this message and exit."),
-       i18n("Print version information."),
-       i18n("Number of seconds to display message for"),
-       i18n("Number of milliseconds to delay per word."),
-       i18n("Set message font (Pango format)."),
-       i18n("Display an image instead of text."),
-       i18n("Display a thought bubble rather than a speech bubble."),
-       i18n("Run xcowsay in daemon mode."),
-       i18n("Size of the cow (small, med, large)."),
-       i18n("Keep daemon attached to terminal."),
-       i18n("Default values for these options can be specified in the "
-            "xcowsay config\nfile. See the man page for more information."),
-       i18n("If the display_time option is not set the display time will "
-            "be calcuated\nfrom the reading_speed parameter multiplied by "
-            "the word count."),
-       i18n("Report bugs to nick@nickg.me.uk")
-       );
+   printf(
+      "%s: xcowsay [OPTION]... [MESSAGE]...\n"
+      "%s\n\n"
+      "%s:\n"
+      " -h, --help\t\t%s\n"
+      " -v, --version\t\t%s\n"
+      " -t, --time=SECONDS\t%s\n"
+      " -r, --reading-speed=N\t%s\n"
+      " -f, --font=FONT\t%s\n"
+      " -d, --dream=FILE\t%s\n"
+      "     --think\t\t%s\n"
+      "     --daemon\t\t%s\n"
+      "     --cow-size=SIZE\t%s\n"
+      "     --image=FILE\t%s\n"
+      "     --monitor=N\t%s\n"
+      "     --at=X,Y\t\t%s\n"
+      "     --bubble-at=X,Y\t%s\n"
+      "     --config=FILE\t%s\n"
+      "     --debug\t\t%s\n\n"
+      "%s\n\n"
+      "%s\n\n"
+      "%s\n",
+      i18n("Usage"),
+      i18n("Display a cow on your desktop with MESSAGE or standard input."),
+      i18n("Options"),
+      i18n("Display this message and exit."),
+      i18n("Print version information."),
+      i18n("Number of seconds to display message for"),
+      i18n("Number of milliseconds to delay per word."),
+      i18n("Set message font (Pango format)."),
+      i18n("Display an image instead of text."),
+      i18n("Display a thought bubble rather than a speech bubble."),
+      i18n("Run xcowsay in daemon mode."),
+      i18n("Size of the cow (small, med, large)."),
+      i18n("Use a different image instead of the cow."),
+      i18n("Display cow on monitor N."),
+      i18n("Force the cow to appear at screen location (X,Y)."),
+      i18n("Change relative position of bubble."),
+      i18n("Specify alternative config file."),
+      i18n("Keep daemon attached to terminal."),
+      i18n("Default values for these options can be specified in the "
+         "xcowsay config\nfile.  See the man page for more information."),
+      i18n("If the display_time option is not set the display time will "
+         "be calcuated\nfrom the reading_speed parameter multiplied by "
+         "the word count.  Set the\ndisplay time to zero to display the "
+         "cow until it is clicked on."),
+      i18n("Report bugs to nick@nickg.me.uk"));
 }
 
 static void version()
 {
    static const char *copy =
-      "Copyright (C) 2008-2009 Nick Gasson\n"
+      "Copyright (C) 2008-2010  Nick Gasson\n"
       "This program comes with ABSOLUTELY NO WARRANTY. This is free software, and\n"
       "you are welcome to redistribute it under certain conditions. See the GNU\n"
       "General Public Licence for details.";
@@ -146,6 +169,30 @@ static int parse_int_option(const char *optarg)
    }
 }
 
+static void parse_position_option(const char *optarg, int *x, int *y)
+{
+   const char *failmsg = i18n("Error: failed to parse '%s' as position\n");
+   
+   char *comma = strchr(optarg, ',');
+   if (comma == NULL) {
+      fprintf(stderr, failmsg, optarg);
+      exit(EXIT_FAILURE);
+   }
+
+   char *endptr;
+   *x = strtol(optarg, &endptr, 10);
+   if (endptr != comma || endptr == optarg) {
+      fprintf(stderr, failmsg, optarg);
+      exit(EXIT_FAILURE);
+   }
+
+   *y = strtol(comma + 1, &endptr, 10);
+   if (*endptr != '\0') {
+      fprintf(stderr, failmsg, optarg);
+      exit(EXIT_FAILURE);
+   }
+}
+
 /*
  * Join all the strings in argv from ind to argc-1 into one big
  * string with spaces between the words.
@@ -186,6 +233,13 @@ int main(int argc, char **argv)
    add_string_option("font", DEF_FONT);
    add_string_option("cow_size", DEF_COW_SIZE);
    add_string_option("image_base", DEF_IMAGE_BASE);
+   add_string_option("alt_image", DEF_ALT_IMAGE);
+   add_int_option("monitor", -1);
+   add_int_option("at_x", -1);
+   add_int_option("at_y", -1);
+   add_int_option("bubble_x", DEF_BUBBLE_X);
+   add_int_option("bubble_y", 0);
+   add_string_option("alt_config_file", "");
    
    parse_config_file();
    
@@ -217,7 +271,33 @@ int main(int argc, char **argv)
          break;
       case 'f':
          set_string_option("font", optarg);
-         break;         
+         break;
+      case 'i':
+         set_string_option("alt_image", optarg);
+         break;
+      case 'm':
+         set_int_option("monitor", parse_int_option(optarg));
+         break;
+      case 'a':
+         {
+            int x, y;
+            parse_position_option(optarg, &x, &y);
+            set_int_option("at_x", x);
+            set_int_option("at_y", y);
+         }
+         break;
+      case 'b':
+         {
+            int x, y;
+            parse_position_option(optarg, &x, &y);
+            set_int_option("bubble_x", x);
+            set_int_option("bubble_y", y);
+         }
+         break;
+      case 'o':
+         set_string_option("alt_config_file", optarg);
+         parse_config_file();
+         break;
       case '?':
          // getopt_long already printed an error message
          failure = 1;
@@ -240,9 +320,23 @@ int main(int argc, char **argv)
       cowsay_init(&argc, &argv);
 
       if (dream_file != NULL) {
-         // COWMODE_DREAM not in daemon yet
-         // TODO: need to make path absolute
-         display_cow_or_invoke_daemon(debug, dream_file, COWMODE_DREAM);
+         // Make path absolute
+         char *wd = getcwd(NULL, 0);
+         char *abs_path;
+
+         asprintf(&abs_path, "%s/%s", wd, dream_file);
+         free(wd);
+
+         struct stat dummy;
+         if (stat(abs_path, &dummy) != 0) {
+            fprintf(stderr, "Error: %s: %s\n", abs_path,
+               strerror(errno));
+            free(abs_path);
+            exit(EXIT_FAILURE);
+         }
+         
+         display_cow_or_invoke_daemon(debug, abs_path, COWMODE_DREAM);
+         free(abs_path);
       }
       else if (optind == argc) {
          read_from_stdin(mode);
